@@ -375,7 +375,7 @@ def load_kernel_config(kernel, custom=False):
     return kernel_config
 
 
-def write_patterns(coords, pattern_name, output_dir, dec=5):
+def write_patterns(coords, pattern_name, output_dir, dec=10):
     """
     Writes coordinates to a text file.
 
@@ -390,12 +390,11 @@ def write_patterns(coords, pattern_name, output_dir, dec=5):
     output_dir : str
         Output path where the file will be saved.
     dec : int
-        Number of decimals to keep in correlation scores.
+        Number of decimals to keep in correlation scores and p-values.
     """
     file_name = pattern_name + ".txt"
     file_path = join(output_dir, file_name)
-    coords.score = np.round(coords.score, dec)
-    coords.to_csv(file_path, sep="\t", index=None)
+    coords.to_csv(file_path, sep="\t", index=None, float_format=f"%.{dec}f")
 
 
 def save_windows(windows, pattern_name, output_dir=".", format="json"):
@@ -498,7 +497,7 @@ def load_bed2d(path):
     return bed2d
 
 
-def download_file(url, file, length=16*1024):
+def download_file(url, file, length=16 * 1024):
     req = urlopen(url)
-    with open(file, 'wb') as fp:
+    with open(file, "wb") as fp:
         shutil.copyfileobj(req, fp, length)
